/*
 * This program uses an algorithm described in:
 *
 *  Multivariate Analysis by K.V. Mardia, J.T. Kent and J.M. Bibby
 *  1979, Academic Press, New York ISBN 0-12-471250-9, p. 397
 *
 * Reference is denoted by : MVA
 *
 * From what I can tell this is a very clear and complete book surveying
 * all the main methods and some of the weird methods of multivariate
 * Analysis.
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include "/usr/local/numerics/recipes_c-kr/misc/nrutil.h"
#include "/usr/local/numerics/recipes_c-kr/misc/nrutil.c"

/*
 * We use MVA Theorem 14.2.1 
 */

char progid[80];

int main(argc,argv)
int argc;
char **argv;
{
  int i,j;
  int     N;            /* Number of Distances */
  double **D;            /* Distance Dmatrix */
  double **A;            /* aij= -0.5 dij^2 */
  double **H;            /* NxN centering Dmatrix: H= I - 1 */
  double **B1,**B;            /* Centered A dmatrix: B=HAH */
  double *DIAG;          /* Diagonal Elements of Householder reduction of B */
  double *ODIAG;         /* Off-Diagonal Elements of same */
  double **ReadD();
  double **MatMult();
  void  tred2(),tqli();
  void PrintDmatrix();
  double dp;
  double sum,sum2,a1,a2;

  strcpy(progid,argv[0]);

  if(argc<2 || argc >3){
    fprintf(stderr,"usage: %s distance_file [design file]\n",progid);
    exit(-1);
  }

  D= (double **) ReadD(argv[1],&N,(argc==3 ? argv[2]: NULL));

/*
  fprintf(stdout,"********** Dmatrix D ************\n");
  PrintDmatrix(D,N,N);
*/

  /* Make A */

  A= (double **) dmatrix(1,N,1,N);

  for(i=1; i<=N; i++)
    for(j=1; j<=N; j++)
      A[i][j]= -0.5*D[i][j]*D[i][j];
/*
  fprintf(stdout,"********** Dmatrix A ************\n");
  PrintDmatrix(A,N,N);
*/
  /* Make H */

  H= (double **) dmatrix(1,N,1,N);

  for(i=1; i<=N; i++)
    for(j=1; j<=N; j++)
      if(i!=j) H[i][j]= -1.0/((double) N);
      else     H[i][j]=  1.0 - 1.0/((double) N);
/*
  fprintf(stdout,"********** Dmatrix H ************\n");
  PrintDmatrix(H,N,N);
*/
  /* Make B */

  B1= (double **) MatMult(A,N,N,H,N,N); /* Right Multiply */
 
/*
  fprintf(stdout,"********** Dmatrix B1 ************\n");
  PrintDmatrix(B1,N,N);
*/
  B= (double **) MatMult(H,N,N,B1,N,N); /* Left  Multiply */

/*
  fprintf(stdout,"********** Dmatrix B ************\n");
  PrintDmatrix(B,N,N);
*/
  /*
   * B is now a centered symmetric dmatrix.
   * We are now going to reduce it to tridiagonal form in order
   * to find its eigenvalues and vectors easily.
   */

  DIAG=  (double *) dvector(1,N);
  ODIAG= (double *) dvector(1,N);
  
  tred2(B,N,DIAG,ODIAG);

  /* 
   * Get the eigenvalues and eigenvectors.
   * B now contains the Q dmatrix from the reduction. 
   */

  tqli(DIAG,ODIAG,N,B);

  /* Print it all out */

  fprintf(stdout,"Eigenvalues:\n");
  sum=0; sum2=0;
  for(i=1; i<=N; i++){ sum += fabs(DIAG[i]); sum2 += DIAG[i]*DIAG[i];}
  a1=0; a2=0;
  for(i=1; i<=N; i++) fprintf(stdout,"\t%d\t%e\t%e\t%e\n",i,DIAG[N+1-i],(a1+= DIAG[N+1-i])/sum,(a2+= DIAG[N+1-i]*DIAG[N+1-i])/sum2);
  fprintf(stdout,"\n");

  /* First Normalize Eigenvectors such that B[i].B[i]= lambda_j */

  for(i=1; i<=N; i++){
    dp=0.0;
    for(j=1; j<=N; j++) dp += (double) B[j][i]*B[j][i];
    for(j=1; j<=N; j++) B[j][i] = B[j][i]*((double) sqrt((double) fabs(DIAG[i])/dp));
  }

  fprintf(stdout,"Eigenvectors:\n\n");
  for(i=1; i<=N; i++) fprintf(stdout,"     %2d     \t",i);
  fprintf(stdout,"\n");
  for(i=N; i>=1; i--){
    for(j=1; j<=N; j++) fprintf(stdout,"%e\t",B[N+1-i][N+1-j]);
    fprintf(stdout,"\n");
  }

exit(0);
}

void PrintDmatrix(M,n,m)
double **M;
int n,m;
{
  int i,j;

  for(i=1; i<=n; i++){
    for(j=1; j<=m; j++) fprintf(stdout,"%e\t",M[i][j]);
    fprintf(stdout,"\n");
  }

}

double **ReadD(name,N,dfile)
char *name,*dfile;
int *N;
{
  int i,j,k,dsgnptr,xp,yp,*dsgnlst;
  FILE *fp,*fp2;
  double **dist;

  if((fp=fopen(name,"r"))==NULL){
    fprintf(stderr,"%s: Unable to find dmatrix file %s\n",progid,name);
    exit(-1);
  }

  fscanf(fp,"%d",N);

  if(dfile==NULL){
    dist= (double **) dmatrix(1,*N,1,*N);
    for(i=1; i<= *N; i++)
      for(j=1; j<= *N; j++)
	fscanf(fp,"%lf",&dist[i][j]);
  } else {

    dsgnlst= (int *) calloc(*N,sizeof(int));
    if((fp2=fopen(dfile,"r"))==NULL){
      fprintf(stderr,"%s: Unable to find design file %s\n",progid,dfile);
      exit(-1);
    }
    
    dsgnptr=0;
    while(fscanf(fp2,"%d",&dsgnlst[dsgnptr])!=EOF) dsgnptr++;
    fclose(fp2);
    dist= (double **) dmatrix(1,dsgnptr,1,dsgnptr);
    xp=yp=1;
    for(i=1; i<= *N; i++){

      for(k=0; k<dsgnptr; k++) if(dsgnlst[k]==i) break;
      if(k==dsgnptr){ /* Discard rest of row */ 
	for(j=1; j<= *N; j++) fscanf(fp,"%*f");
	continue;
      }
      yp=1;
      for(j=1; j<= *N; j++){
	
	for(k=0; k<dsgnptr; k++) if(dsgnlst[k]==j) break;
	if(k==dsgnptr){fscanf(fp,"%*f");  continue;} /* Skip Column */
	
	fscanf(fp,"%lf",&dist[xp][yp]);
	yp++;
      }
      xp++;
    }
    *N=dsgnptr;
  }

  fclose(fp);

  return(dist);
}

double **MatMult(M1,r,s,M2,p,q)
double **M1,**M2;
int r,s,p,q;
{
  register int j,l,k;
  double **M;

  if(s!=p) {
    fprintf(stderr,"MatMult: Error! Inner Indices of Matrices not equal (%d,%d). Exiting\n",s,p);
    exit(-1);
  }

  M= (double **) dmatrix(1,r,1,q);
  for(l=1; l<=r; l++)
   for(k=1; k<=q; k++){
     M[l][k]=0.0;
     for(j=1; j<=p; j++) M[l][k] += M1[l][j]*M2[j][k];
   }

return(M);
}

void tred2(a,n,d,e)
double **a,d[],e[];
int n;
{
	int l,k,j,i;
	double scale,hh,h,g,f;

	for (i=n;i>=2;i--) {
		l=i-1;
		h=scale=0.0;
		if (l > 1) {
			for (k=1;k<=l;k++)
				scale += fabs(a[i][k]);
			if (scale == 0.0)
				e[i]=a[i][l];
			else {
				for (k=1;k<=l;k++) {
					a[i][k] /= scale;
					h += a[i][k]*a[i][k];
				}
				f=a[i][l];
				g = f>0 ? -sqrt(h) : sqrt(h);
				e[i]=scale*g;
				h -= f*g;
				a[i][l]=f-g;
				f=0.0;
				for (j=1;j<=l;j++) {
				/* Next statement can be omitted if eigenvectors not wanted */
					a[j][i]=a[i][j]/h;
					g=0.0;
					for (k=1;k<=j;k++)
						g += a[j][k]*a[i][k];
					for (k=j+1;k<=l;k++)
						g += a[k][j]*a[i][k];
					e[j]=g/h;
					f += e[j]*a[i][j];
				}
				hh=f/(h+h);
				for (j=1;j<=l;j++) {
					f=a[i][j];
					e[j]=g=e[j]-hh*f;
					for (k=1;k<=j;k++)
						a[j][k] -= (f*e[k]+g*a[i][k]);
				}
			}
		} else
			e[i]=a[i][l];
		d[i]=h;
	}
	/* Next statement can be omitted if eigenvectors not wanted */
	d[1]=0.0;
	e[1]=0.0;
	/* Contents of this loop can be omitted if eigenvectors not
			wanted except for statement d[i]=a[i][i]; */
	for (i=1;i<=n;i++) {
		l=i-1;
		if (d[i]) {
			for (j=1;j<=l;j++) {
				g=0.0;
				for (k=1;k<=l;k++)
					g += a[i][k]*a[k][j];
				for (k=1;k<=l;k++)
					a[k][j] -= g*a[k][i];
			}
		}
		d[i]=a[i][i];
		a[i][i]=1.0;
		for (j=1;j<=l;j++) a[j][i]=a[i][j]=0.0;
	}
}

#define ASIGN(a,b) ((b)<0 ? -fabs(a) : fabs(a))

void tqli(d,e,n,z)
double d[],e[],**z;
int n;
{
	int m,l,iter,i,k;
	double s,r,p,g,f,dd,c,b;
	void nrerror();

	for (i=2;i<=n;i++) e[i-1]=e[i];
	e[n]=0.0;
	for (l=1;l<=n;l++) {
		iter=0;
		do {
			for (m=l;m<=n-1;m++) {
				dd=fabs(d[m])+fabs(d[m+1]);
				if (fabs(e[m])+dd == dd) break;
			}
			if (m != l) {
				if (iter++ == 30) nrerror("Too many iterations in TQLI");
				g=(d[l+1]-d[l])/(2.0*e[l]);
				r=sqrt((g*g)+1.0);
				g=d[m]-d[l]+e[l]/(g+ASIGN(r,g));
				s=c=1.0;
				p=0.0;
				for (i=m-1;i>=l;i--) {
					f=s*e[i];
					b=c*e[i];
					if (fabs(f) >= fabs(g)) {
						c=g/f;
						r=sqrt((c*c)+1.0);
						e[i+1]=f*r;
						c *= (s=1.0/r);
					} else {
						s=f/g;
						r=sqrt((s*s)+1.0);
						e[i+1]=g*r;
						s *= (c=1.0/r);
					}
					g=d[i+1]-p;
					r=(d[i]-g)*s+2.0*c*b;
					p=s*r;
					d[i+1]=g+p;
					g=c*r-b;
					/* Next loop can be omitted if eigenvectors not wanted */
					for (k=1;k<=n;k++) {
						f=z[k][i+1];
						z[k][i+1]=s*z[k][i]+c*f;
						z[k][i]=c*z[k][i]-s*f;
					}
				}
				d[l]=d[l]-p;
				e[l]=g;
				e[m]=0.0;
			}
		} while (m != l);
	}
}







